// Copyright 2003-2007 Tom Felker, wore, wiesl
//
// VLevel foobar2000 plugin
//
// VLevel is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// VLevel is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with VLevel; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

// foo_dsp_vlevel.cpp - the foobar2000 plugin, uses VolumeLeveler
// based on vlevel-bin.cpp
// wore <info@wore.ma.cx>
//
// Crash fixes, dark mode and 64-bit support by Janne H.

#include "../foobar2000/helpers/foobar2000+atl.h"
#include "../foobar2000/SDK/coreDarkMode.h"
#include "../libPPUI/wtl-pp.h"
#include <math.h>
#include <float.h>
#include "volumeleveler.h"
#include "resource.h"

#ifdef _DEBUG
#define COMPONENT_VERSION_NAME "VLevel - DEBUG BUILD!"
#else
#define COMPONENT_VERSION_NAME "VLevel"
#endif
DECLARE_COMPONENT_VERSION(COMPONENT_VERSION_NAME, "20230827",
    "volume levelling plugin\n\n"
    "made ready for 0.9.x series, config settings now work well, added useful limits for max multiplier, added dB scale, added debug code by wiesl\n\n"
    "written by wore\n"
    "additional code by Tom Felker and ssamadhi97\n"
    "ported to foobar2000 0.9 by G-Lite\n\n"
    "based on VLevel 0.5 by Tom Felker (http://vlevel.sourceforge.net/news/)"
);

//#define FOO_DSP_VLEVEL_DEBUG

// {EC001F79-9D79-4dc8-B7CB-40818D7A1009}
static const GUID g_foo_dsp_vlevel_guid = { 0xec001f79, 0x9d79, 0x4dc8, { 0xb7, 0xcb, 0x40, 0x81, 0x8d, 0x7a, 0x10, 0x9 } };

#define snprintf _snprintf

// http://en.wikipedia.org/wiki/Variadic_macro
// http://en.wikipedia.org/wiki/C_preprocessor
// http://www.geocities.com/davinpearson/research/1999/sm4dcp.html

//      DEBUG("on_playlist_created begin (\"%s\", %li)", p_name, p_index)
#ifdef FOO_DSP_VLEVEL_DEBUG
#define DEBUG(format, ...)                                       \
    do                                                           \
    {                                                            \
        MyDebug::fprintf_flush((format), __VA_ARGS__);             \
    } while (0);
#else
#define DEBUG(format, ...) ((void)0)
#endif

#ifdef FOO_DSP_VLEVEL_DEBUG
  #define DEBUG_VLEVEL_INFO(prefix, var) DEBUG("%s: Using Strength=%lf, MaxMultiplier=%li, BufferLength=%li", (prefix), (double)((var).strength) / 100, (long)((var).max_multiplier), (long)((var).buffer_length));
#else
  #define DEBUG_VLEVEL_INFO(prefix, var) ((void)0)
#endif

#ifdef FOO_DSP_VLEVEL_DEBUG
class MyDebug
{
protected:
    static FILE* log;

public:
    MyDebug()
    {
        log = fopen("C:\\foobar2000.log", "w");
        DEBUG("opened log");
    }
    virtual ~MyDebug()
    {
        DEBUG("closed log");
        (void)fclose(log);
    }

    static void fprintf_flush(const char* format, ...)
    {
        // http://www.ddj.com/cpp/184403430
        // http://gcc.gnu.org/ml/gcc-bugs/2001-09/msg00147.html
        // http://www.eskimo.com/~scs/cclass/int/sx11c.html
        // http://lists.mysql.com/internals/2180?f=plain

        va_list va_args;
        va_start(va_args, format);
        vfprintf(log, format, va_args);
        va_end(va_args);

        fprintf(log, "\n");

        flush();
    }

    static void flush()
    {
        fflush(log);
    }
};

FILE* MyDebug::log = NULL;
static MyDebug static_MyDebug;
#endif

struct vlevel_preset_info {
    t_int8 m_strength;
    t_int8 m_max_multiplier;
    t_int8 m_buffer_length;

    bool get_data(const dsp_preset &p_data)
    {
        DEBUG("vlevel_preset_info::get_data() entry");

        DEBUG("vlevel_preset_info::datasize=%li, should be=%li", (long)p_data.get_data_size(), sizeof(vlevel_preset_info));

        if (p_data.get_data_size() != sizeof(vlevel_preset_info)) return false;
        const vlevel_preset_info temp = *(const vlevel_preset_info *)p_data.get_data();
        m_strength = temp.m_strength;
        m_max_multiplier = temp.m_max_multiplier;
        m_buffer_length = temp.m_buffer_length;

        DEBUG_VLEVEL_INFO("vlevel_preset_info::get_data()", *this);

        DEBUG("vlevel_preset_info::get_data() return");
        return true;
    }

    bool set_data(dsp_preset &p_data)
    {
        DEBUG("vlevel_preset_info::set_data() entry");

        DEBUG_VLEVEL_INFO("vlevel_preset_info::set_data()", *this);

        vlevel_preset_info temp = {m_strength, m_max_multiplier, m_buffer_length};
        p_data.set_data(&temp, sizeof(temp));

        DEBUG("vlevel_preset_info::set_data() return");

        return true;
    }
};

static void RunDSPConfigPopup(const dsp_preset &p_data, HWND p_parent, dsp_preset_edit_callback &p_callback);

class dialog_vlevel_config : public CDialogImpl<dialog_vlevel_config> {
    public:
    dialog_vlevel_config(const dsp_preset &initData, dsp_preset_edit_callback &callback) : m_callback(callback)
    {
        m_info.get_data(initData);
        DEBUG_VLEVEL_INFO("dialog_vlevel_config::dialog_vlevel_config", p_data);
    }

    enum { IDD = IDD_CONFIG };

    BEGIN_MSG_MAP(dialog_vlevel_config)
    MSG_WM_INITDIALOG(OnInitDialog)
    COMMAND_HANDLER_EX(IDOK, BN_CLICKED, OnButton)
    COMMAND_HANDLER_EX(IDCANCEL, BN_CLICKED, OnButton)
    MSG_WM_HSCROLL(OnHScroll)
    END_MSG_MAP()

private:
    BOOL OnInitDialog(CWindow, LPARAM)
    {
        HWND slider;

        m_hooks.AddDialogWithControls(m_hWnd);

        DEBUG_VLEVEL_INFO("dialog_vlevel_config::on_message(WM_INITDIALOG)", m_info);

        // allow strength slider settings of 0-100 (strength*100) so users can set this parameter with 0.01 granularity
        slider = GetDlgItem(IDC_STRENGTH);
        uSendMessage(slider, TBM_SETRANGE, 0, MAKELONG(0, 100));
        uSendMessage(slider, TBM_SETPOS, 1, m_info.m_strength);

        slider = GetDlgItem(IDC_MAX_MULTIPLIER);
        uSendMessage(slider, TBM_SETRANGE, 0, MAKELONG(1, 100)); // GW: Changed from 1-40 instead of 0-40 (so MAX_HUGE can not be set!)
        uSendMessage(slider, TBM_SETPOS, 1, m_info.m_max_multiplier);

        // same idea as strength, slider from 1-50, which comes down to 0.1 to 5.0 seconds
        slider = GetDlgItem(IDC_BUFFER_LENGTH);
        uSendMessage(slider, TBM_SETRANGE, 0, MAKELONG(1, 50));
        uSendMessage(slider, TBM_SETPOS, 1, m_info.m_buffer_length);

        update_display();
        return TRUE;
    }

    void OnHScroll(UINT nSBCode, UINT nPos, CScrollBar pScrollBar)
    {
        //vlevel_preset_info info = { 80, 25, 20 };
        m_info.m_strength = (t_int8)uSendDlgItemMessage(IDC_STRENGTH, TBM_GETPOS, 0, 0);
        m_info.m_max_multiplier = (t_int8)uSendDlgItemMessage(IDC_MAX_MULTIPLIER, TBM_GETPOS, 0, 0);
        m_info.m_buffer_length = (t_int8)uSendDlgItemMessage(IDC_BUFFER_LENGTH, TBM_GETPOS, 0, 0);
        dsp_preset_impl preset;
        preset.set_owner(g_foo_dsp_vlevel_guid);
        m_info.set_data(preset);

        m_callback.on_preset_changed(preset);

        update_display();
    }

    void OnButton(UINT, int id, CWindow)
    {
        EndDialog(id);
    }

    void update_display() // updates textual display of slider values
    {
        char temp[128];

        // convert strength slider position to display the actual strength value VLevel will receive later on
        sprintf(temp, "%1.2f", ((double)uSendDlgItemMessage(IDC_STRENGTH, TBM_GETPOS, 0, 0)) / 100);
        uSetDlgItemText(m_hWnd, IDC_STATIC_STRENGTH, temp);

        LRESULT maxmultiplier = uSendDlgItemMessage(IDC_MAX_MULTIPLIER, TBM_GETPOS, 0, 0);
        sprintf(temp, "%d (%+4.1lf dB)", (int)maxmultiplier, 20*log10((double)maxmultiplier));
        uSetDlgItemText(m_hWnd, IDC_STATIC_MAX_MULTIPLIER, temp);

        sprintf(temp, "%1.1f s", ((double)uSendDlgItemMessage(IDC_BUFFER_LENGTH, TBM_GETPOS, 0, 0)) / 10);
        uSetDlgItemText(m_hWnd, IDC_STATIC_BUFFER_LENGTH, temp);
    }

    vlevel_preset_info m_info;
    dsp_preset_edit_callback &m_callback;
    fb2k::CCoreDarkModeHooks m_hooks;
};

class dsp_vlevel : public dsp_impl_base {
private:
    VolumeLeveler *m_vl;

    // always == vl->GetChannels()
    size_t m_channels;

    // only applies to our local buffers which are cached across on_chunk() calls
    size_t m_samples;

    // our local buffers
    value_t *m_raw_value_buf;
    value_t **m_bufs;

    // only used to SetLengthAndChannels and to create chunks
    UINT m_srate;
    UINT m_chc;

    t_int8 m_buffer_length;

    // make sure GUI resets the flag using the update_configuration method
    bool m_buffer_length_changed;

    bool ready()
    {
        // could be optimized to vl&&bufs, or expanded for complete sanity checks
        return m_vl && m_bufs && m_raw_value_buf && m_channels == m_vl->GetChannels();
    }

    // insert the contents of our buffers into the stream
    // can only be called from on_endoftrack(), on_endofplayback(), and on_chunk()
    void flush_data()
    {
        // foobar2000 sometimes calls this when it shouldn't, so this is necessary.
        if (!ready()) return;

        // do nothing if there's no audio in the vl
        // otherwise on_chunk() will make this chunk empty, and foobar2000 complains
        // next vl version will have GetGoodSamples(){return samples - silence};
        if (m_vl->GetSamples() - m_vl->GetSilence() == 0) return;

        // make a new chunk filled with enough silence to flush the buffers.
        if (m_vl->GetSamples() > 0) {
            audio_chunk *chunk = insert_chunk(m_vl->GetSamples() * m_channels);
            if (!chunk) throw std::bad_alloc(); // shouldn't be needed
            t_size new_size = m_vl->GetSamples() * m_channels;
            chunk->set_data_size(new_size);
            chunk->set_channels((unsigned int)m_channels, m_chc);
            chunk->set_srate(m_srate);
            chunk->set_silence(m_vl->GetSamples());

            // yes, it's that simple.
            on_chunk(chunk);
        }
        m_vl->Flush();
    }

    // only affects the local buffers, not the volumeleveler
    void allocate_buffers(size_t new_samples, size_t new_channels)
    {
        cleanup_buffers();
        m_samples = new_samples;
        m_channels = new_channels;

        if (m_samples > 0 && m_channels > 0) {
            m_raw_value_buf = new value_t[m_samples * m_channels];
            m_bufs = new value_t * [m_channels];
            for (size_t ch = 0; ch < m_channels; ++ch) {
                m_bufs[ch] = new value_t[m_samples];
            }
        }
    }

    // only affects the local buffers, not the volumeleveler
    void cleanup_buffers(void)
    {
        if (m_raw_value_buf) {
            delete[] m_raw_value_buf;
            m_raw_value_buf = NULL;
        }
        if (m_bufs) {
            for (size_t ch = 0; ch < m_channels; ++ch) {
                delete[] m_bufs[ch];
            }
            delete[] m_bufs;
            m_bufs = NULL;
        }
        m_samples = 0;
        m_channels = 0;
    }

    bool on_chunk(audio_chunk *chunk)
    {
        if (chunk->get_sample_count() == 0 || chunk->get_sample_rate() == 0 || chunk->get_channel_count() == 0 || chunk->get_data() == NULL) return false;

        bool reinit_vl = false;
        // we must change the vl's settings if buffer length or channels changes
        if (chunk->get_srate() != m_srate || chunk->get_channels() != m_channels || m_buffer_length_changed) {
            flush_data();
            reinit_vl = true;
        }

        // if this chunk is a different size than last time, we must get new temp buffers
        if (chunk->get_sample_count() != m_samples || chunk->get_channels() != m_channels) {
            allocate_buffers(chunk->get_sample_count(), chunk->get_channels());
        }

        if (reinit_vl) {
            m_srate = chunk->get_srate();
            m_channels = chunk->get_channels();
            m_chc = chunk->get_channel_config();
            m_vl->SetSamplesAndChannels((size_t)(((double)m_buffer_length) / 10.0 * m_srate + 0.5), m_channels);
            m_buffer_length_changed = false;
        }

        if (!ready()) return true;

        audio_sample *data = chunk->get_data();

        // de-interleave the data
        for (size_t s = 0; s < m_samples; ++s) {
            for (size_t ch = 0; ch < m_channels; ++ch) {
                m_bufs[ch][s] = data[s * m_channels + ch];
            }
        }

        // do the exchange
        size_t silence_samples = m_vl->Exchange(m_bufs, m_bufs, m_samples);

        // interleave the data
        for (size_t s = silence_samples; s < m_samples; ++s) {
            for (size_t ch = 0; ch < m_channels; ++ch) {
                data[(s - silence_samples) * m_channels + ch] = m_bufs[ch][s];
            }
        }

        // set the amount of good samples in the chunk
        chunk->set_sample_count(m_samples - silence_samples);

        return (m_samples - silence_samples > 0);
    }

public:
    dsp_vlevel(const dsp_preset &p_data) : m_vl(NULL), m_raw_value_buf(NULL), m_bufs(NULL), m_channels(0), m_samples(0), m_srate(0), m_chc(0), m_buffer_length_changed(false)
    {
        DEBUG("dsp_vlevel::dsp_vlevel() entry");

        m_buffer_length = 20;
        value_t strength = 0.8;
        value_t maxmultiplier = 25;

        vlevel_preset_info info;
        dsp_preset_impl temp = p_data;
        if (info.get_data(temp)) {
            DEBUG("dsp_vlevel::dsp_vlevel() using NON default values");
            DEBUG_VLEVEL_INFO("dialog_vlevel_config::on_message(IDCANCEL)", info);
            strength = (value_t)(((double)info.m_strength) / 100); // must be divided by 100!!
            maxmultiplier = info.m_max_multiplier;
            m_buffer_length = info.m_buffer_length;
        } else {
            DEBUG("dsp_vlevel::dsp_vlevel() using default values");
        }

        m_vl = new VolumeLeveler();
        m_vl->SetStrength(strength);
        m_vl->SetMaxMultiplier(maxmultiplier);

        DEBUG("dsp_vlevel::dsp_vlevel() return");
    }

    ~dsp_vlevel()
    {
        DEBUG("dsp_vlevel::~dsp_vlevel() entry");

        cleanup_buffers();
        delete m_vl;

        DEBUG("dsp_vlevel::~dsp_vlevel() return");
    }

    static GUID g_get_guid()
    {
        DEBUG("dsp_vlevel::g_get_guid() entry");
        return g_foo_dsp_vlevel_guid;
    }

    static void g_get_name(pfc::string_base &p_out)
    {
        DEBUG("dsp_vlevel::g_get_name() entry");
        p_out = "VLevel";
    }

    static bool g_get_default_preset(dsp_preset &p_out)
    {
        DEBUG("dsp_vlevel::g_get_default_preset() entry");
        vlevel_preset_info info = {80, 25, 20};
        p_out.set_owner(g_get_guid());
        info.set_data(p_out);
        DEBUG("dsp_vlevel::g_get_default_preset() return");
        return true;
    }

    static bool g_have_config_popup()
    {
        DEBUG("dsp_vlevel::g_have_config_popup() entry");
        return true;
    }

    static void g_show_config_popup(const dsp_preset &p_data, HWND p_parent, dsp_preset_edit_callback &p_callback)
    {
        DEBUG("dsp_vlevel::g_show_config_popup() entry");
        ::RunDSPConfigPopup(p_data, p_parent, p_callback);
    }

    bool set_data(const dsp_preset &p_data)
    {
        DEBUG("dsp_vlevel::set_data() entry");
        vlevel_preset_info info;
        if (!info.get_data(p_data)) return false;

        DEBUG_VLEVEL_INFO("dsp_vlevel::set_data()", info);

        m_vl->SetStrength((value_t)((double)info.m_strength / 100));
        m_vl->SetMaxMultiplier(info.m_max_multiplier);
        m_buffer_length = info.m_buffer_length;
        m_buffer_length_changed = true;

        DEBUG("dsp_vlevel::set_data() return");
        return true;
    }

    bool on_chunk(audio_chunk * chunk, foobar2000_io::abort_callback &p_abort)
    {
        DEBUG("dsp_vlevel::on_chunk() entry");
        bool ret = on_chunk(chunk);

        DEBUG("dsp_vlevel::on_chunk() return");
        return ret;
    }

    // this is redundant if on_endoftrack is always called first.
    void on_endofplayback(foobar2000_io::abort_callback &p_abort) override
    {
        DEBUG("dsp_vlevel::on_endofplayback() entry");

        //console::info("on_endofplayback");
        flush_data();

        DEBUG("dsp_vlevel::on_endofplayback() return");
    }

    // we output the contents of our buffers
    void on_endoftrack(foobar2000_io::abort_callback &p_abort) override
    {
        DEBUG("dsp_vlevel::on_endoftrack() entry");

        //console::info("on_endoftrack");
        flush_data();

        DEBUG("dsp_vlevel::on_endofplayback() return");
    }

    // called on seeks, so we drop our data
    void flush() override
    {
        DEBUG("dsp_vlevel::flush() entry");

        //console::info("flush");
        m_vl->Flush();

        DEBUG("dsp_vlevel::flush() return");
    }

    // this is called very often while playing
    double get_latency() override
    {
        DEBUG("dsp_vlevel::get_latency()");
        return (m_srate == 0) ? 0 : ((double)(m_vl->GetSamples() - m_vl->GetSilence()) / m_srate);
    }

    bool need_track_change_mark() override
    {
        DEBUG("dsp_vlevel::need_track_change_mark()");
        return false;
    }
};

static void RunDSPConfigPopup(const dsp_preset &p_data, HWND p_parent, dsp_preset_edit_callback &p_callback)
{
    dialog_vlevel_config popup(p_data, p_callback);
    if (popup.DoModal(p_parent) != IDOK) p_callback.on_preset_changed(p_data);
}

static dsp_factory_t<dsp_vlevel> g_foo;
